# -*- encoding: utf-8 -*-
"""Defines a set of mixins that provide tools for interacting with entities."""
from fauxfactory import gen_choice
from nailgun import client, config
from nailgun.entity_fields import (
    Field,
    _get_class,
    IntegerField,
    OneToManyField,
    OneToOneField,
)
import threading
import time

from sys import version_info
if version_info.major == 2:
    # pylint:disable=import-error
    from urlparse import urljoin
    import httplib as http_client
    import thread
else:
    from urllib.parse import urljoin  # pylint:disable=F0401,E0611
    import _thread as thread  # pylint:disable=import-error
    import http.client as http_client  # pylint:disable=import-error


#: Default for ``poll_rate`` argument to
#: :func:`nailgun.entity_mixins._poll_task`.
TASK_POLL_RATE = 5
#: Default for ``timeout`` argument to
#: :func:`nailgun.entity_mixins._poll_task`.
TASK_TIMEOUT = 300

#: A :class:`nailgun.config.ServerConfig` object.
#:
#: Used by :class:`nailgun.entity_mixins.Entity`.
DEFAULT_SERVER_CONFIG = None


class TaskTimeout(Exception):
    """Indicates that a task did not finish before the timout limit."""


def _poll_task(task_id, server_config, poll_rate=None, timeout=None):
    """Implement ``robottelo.entities.ForemanTask.poll``.

    See ``robottelo.entities.ForemanTask.poll`` for a full description of how
    this method acts. Other methods may also call this method, such as
    :meth:`nailgun.entity_mixins.EntityDeleteMixin.delete`.

    This function has been placed in this module to keep the import tree sane.
    This function could also be placed in ``robottelo.api.utils``. However,
    doing so precludes ``robottelo.api.utils`` from importing
    ``robottelo.entities``, which may be desirable in the future.

    This function is private because only entity mixins should use this.
    ``robottelo.entities.ForemanTask`` is, for obvious reasons, an exception.

    """
    if poll_rate is None:
        poll_rate = TASK_POLL_RATE
    if timeout is None:
        timeout = TASK_TIMEOUT

    # Implement the timeout.
    def raise_task_timeout():
        """Raise a KeyboardInterrupt exception in the main thread."""
        thread.interrupt_main()
    timer = threading.Timer(timeout, raise_task_timeout)

    # Poll until the task finishes. The timeout prevents an infinite loop.
    try:
        timer.start()
        path = '{0}/foreman_tasks/api/tasks/{1}'.format(
            server_config.url,
            task_id
        )
        while True:
            response = client.get(
                path,
                auth=server_config.auth,
                verify=server_config.verify,
            )
            response.raise_for_status()
            task_info = response.json()
            if task_info['state'] != 'running':
                return task_info
            time.sleep(poll_rate)
    except KeyboardInterrupt:
        # raise_task_timeout will raise a KeyboardInterrupt when the timeout
        # expires. Catch the exception and raise TaskTimeout
        raise TaskTimeout("Timed out polling task {0}".format(task_id))
    finally:
        timer.cancel()


def _make_entity_from_id(entity_cls, entity_obj_or_id, server_config):
    """Given an entity object or an ID, return an entity object.

    If the value passed in is an object that is a subclass of class ``Entity``,
    return that value. Otherwise, create an object of the type that ``field``
    references, give that object an ID of ``field_value``, and return that
    object.

    :param entity_cls: An :class:`Entity` subclass.
    :param entity_obj_or_id: Either a :class:`nailgun.entity_mixins.Entity`
        object or an entity ID.
    :returns: An ``entity_cls`` object.
    :rtype: nailgun.entity_mixins.Entity

    """
    if isinstance(entity_obj_or_id, entity_cls):
        return entity_obj_or_id
    return entity_cls(server_config, id=entity_obj_or_id)


def _make_entities_from_ids(entity_cls, entity_objs_and_ids, server_config):
    """Given an iterable of entities and/or IDs, return a list of entities.

    :param entity_cls: An :class:`Entity` subclass.
    :param entity_obj_or_id: An iterable of
        :class:`nailgun.entity_mixins.Entity` objects and/or entity IDs.
    :returns: A list of ``entity_cls`` objects.

    """
    return [
        _make_entity_from_id(entity_cls, entity_or_id, server_config)
        for entity_or_id
        in entity_objs_and_ids
    ]


# -----------------------------------------------------------------------------
# Definition of parent Entity class and its dependencies.
# -----------------------------------------------------------------------------


class NoSuchPathError(Exception):
    """Indicates that the requested path cannot be constructed."""


class NoSuchFieldError(Exception):
    """Indicates that the assigned-to :class:`Entity` field does not exist."""


class Entity(object):
    """A logical representation of a Foreman entity.

    This class is rather useless as is, and it is intended to be subclassed.
    Subclasses can specify two useful types of information:

    * fields
    * metadata

    Fields are represented by setting class attributes, and metadata is
    represented by settings attributes on the inner class named ``Meta``. For
    example, consider this class declaration::

        class User(Entity):
            name = StringField()
            supervisor = OneToOneField('User')
            subordinate = OneToManyField('User')

            class Meta(object):
                api_path = 'api/users'

    In the example above, the class attributes of ``User`` are fields, and the
    class attributes of ``Meta`` are metadata. Here is one way to instantiate
    the ``User`` object shown above::

        User(
            name='Alice',
            supervisor=User(id=1),
            subordinate=[User(id=3), User(id=4)],
        )

    The canonical procedure for initializing foreign key fields, shown above,
    is powerful but verbose. As an alternative, the following convenience is
    offered:

        User(name='Alice', supervisor=1, subordinate=[3, 4])

    An entity object is useless if you are unable to use it to communicate with
    a server. The solution is to provide a ``nailgun.config.ServerConfig`` when
    instantiating a new entity. This configuration object is stored as an
    instance variable named ``_server_config`` and used by methods such as
    :meth:`nailgun.entity_mixins.Entity.path`.

    1. If the ``server_config`` argument is specified, then that is used.
    2. Otherwise, if :data:`nailgun.entity_mixins.DEFAULT_SERVER_CONFIG` is
       set, then that is used.
    3. Otherwise, call ``nailgun.config.ServerConfig.get()``.

    """
    # The id() builtin is still available within instance methods, class
    # methods, static methods, inner classes, and so on. However, id() is *not*
    # available at the current level of lexical scoping after this point.
    id = IntegerField()  # pylint:disable=C0103

    def __init__(self, server_config=None, **kwargs):
        # server_config > DEFAULT_SERVER_CONFIG > ServerConfig.get()
        if server_config is not None:
            self._server_config = server_config
        elif DEFAULT_SERVER_CONFIG is not None:
            self._server_config = DEFAULT_SERVER_CONFIG
        else:
            self._server_config = config.ServerConfig.get()

        # Check that a valid set of field values has been passed in.
        fields = self.get_fields()
        if not set(kwargs.keys()).issubset(fields.keys()):
            raise NoSuchFieldError(
                'Valid fields are {0}, but received {1} instead.'.format(
                    fields.keys(), kwargs.keys()
                )
            )

        # Iterate through the values passed in and assign them as instance
        # variable to `self`. Make sure to transform entity IDs into entity
        # objects. (This feature is described in the docstring.)
        for field_name, field_value in kwargs.items():  # e.g. ('admin', True)
            field = fields[field_name]  # e.g. A BooleanField object
            if isinstance(field, OneToOneField):
                setattr(self, field_name, _make_entity_from_id(
                    _get_class(field.entity, 'robottelo.entities'),
                    field_value,
                    self._server_config
                ))
            elif isinstance(field, OneToManyField):
                setattr(self, field_name, _make_entities_from_ids(
                    _get_class(field.entity, 'robottelo.entities'),
                    field_value,
                    self._server_config
                ))
            else:
                setattr(self, field_name, field_value)

    class Meta(object):  # (too-few-public-methods) pylint:disable=R0903
        """Non-field information about this entity.

        This class is a convenient place to store any non-field information
        about an entity. For example, the ``api_path`` variable is used by
        :meth:`nailgun.entity_mixins.Entity.path`.

        """

    def path(self, which=None):
        """Return the path to the current entity.

        Return the path to base entities of this entity's type if:

        * ``which`` is ``'base'``, or
        * ``which`` is ``None`` and instance attribute ``id`` is unset.

        Return the path to this exact entity if instance attribute ``id`` is
        set and:

        * ``which`` is ``'self'``, or
        * ``which`` is ``None``.

        Raise :class:`NoSuchPathError` otherwise.

        Child classes may choose to extend this method, especially if a child
        entity offers more than the two URLs supported by default. If extended,
        then the extending class should check for custom parameters before
        calling ``super``::

            def path(self, which):
                if which == 'custom':
                    return urljoin(…)
                super(ChildEntity, self).__init__(which)

        This will allow the extending method to accept a custom parameter
        without accidentally raising a :class:`NoSuchPathError`.

        :param which: A string. Optional. Valid arguments are 'self' and
            'base'.
        :return: A string. A fully qualified URL.
        :raises nailgun.entity_mixins.NoSuchPathError: If no path can be built.

        """
        # It is OK that member ``self.Meta.api_path`` is not found. Subclasses
        # are required to set that attribute if they wish to use this method.
        #
        # Beware of leading and trailing slashes:
        #
        # urljoin('example.com', 'foo') => 'foo'
        # urljoin('example.com/', 'foo') => 'example.com/foo'
        # urljoin('example.com', '/foo') => '/foo'
        # urljoin('example.com/', '/foo') => '/foo'
        base = urljoin(
            self._server_config.url + '/',
            self.Meta.api_path  # pylint:disable=no-member
        )
        if which == 'base' or (which is None and 'id' not in vars(self)):
            return base
        elif (which == 'self' or which is None) and 'id' in vars(self):
            return urljoin(base + '/', str(self.id))
        raise NoSuchPathError

    @classmethod
    def get_fields(cls):
        """Return all fields defined on the current class.

        :return: A dict mapping class attribute names to ``Field`` objects.

        """
        # When `dir` is called on "a type or class object, the list contains
        # the names of its attributes, and recursively of the attributes of its
        # bases." In constrast, `vars(cls)` returns only the attributes of
        # `cls` while ignoring all parent classes. Thus, this fails for child
        # classes:
        #
        #     for key, val in vars(cls).items():
        #         if isinstance(val, Field):
        #             attrs[key] = val
        #
        attrs = {}
        for field_name in dir(cls):
            field = getattr(cls, field_name)
            if isinstance(field, Field):
                attrs[field_name] = field
        return attrs

    def get_values(self):
        """Return the value of each field on the current object.

        This method is almost identical to ``vars(self).copy()``. However, only
        instance attributes that correspond to a field are included in the
        returned dict.

        :return: A dict mapping class attribute names to field values.

        """
        attrs = vars(self).copy()
        attrs.pop('_server_config')
        return attrs


class EntityDeleteMixin(object):
    """A mixin that adds the ability to delete an entity."""

    def delete_raw(self):
        """Delete the current entity.

        Send an HTTP DELETE request to :meth:`Entity.path`. Return the
        response. Do not check the response for any errors, such as an HTTP 4XX
        or 5XX status code.

        """
        return client.delete(
            self.path(which='self'),
            auth=self._server_config.auth,
            verify=self._server_config.verify,
        )

    def delete(self, synchronous=True):
        """Delete the current entity.

        Call :meth:`delete_raw` and check for an HTTP 4XX or 5XX response.
        Return either the JSON-decoded response or information about a
        completed foreman task.

        :param synchronous: A boolean. What should happen if the server returns
            an HTTP 202 (accepted) status code? Wait for the task to complete
            if ``True``. Immediately return a response otherwise.
        :returns: A dict. Either the JSON-decoded response or information about
            a foreman task.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If an HTTP 202 response is received and the
            response JSON can not be decoded.
        :raises nailgun.entity_mixins.TaskTimeout: If an HTTP 202 response is
            received, ``synchronous is True`` and the task times out.

        """
        response = self.delete_raw()
        response.raise_for_status()
        if (synchronous is True and
                response.status_code == http_client.ACCEPTED):
            return _poll_task(response.json()['id'], self._server_config)
        if response.status_code == http_client.NO_CONTENT:
            # "The server successfully processed the request, but is not
            # returning any content. Usually used as a response to a successful
            # delete request."
            return
        return response.json()


class EntityReadMixin(object):
    """A mixin that provides the ability to read an entity."""

    def read_raw(self):
        """Get information about the current entity.

        Send an HTTP GET request to :meth:`Entity.path`. Return the response.
        Do not check the response for any errors, such as an HTTP 4XX or 5XX
        status code.

        :return: A ``requests.response`` object.

        """
        return client.get(
            self.path('self'),
            auth=self._server_config.auth,
            verify=self._server_config.verify,
        )

    def read_json(self):
        """Get information about the current entity.

        Call :meth:`read_raw`. Check the response status code, decode JSON and
        return the decoded JSON as a dict.

        :return: A dict. The server's response, with all JSON decoded.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If the response JSON can not be decoded.

        """
        response = self.read_raw()
        response.raise_for_status()
        return response.json()

    def read(self, entity=None, attrs=None, ignore=()):
        """Get information about the current entity.

        Call :meth:`read_json`. Use this information to populate an object of
        type ``type(self)`` and return that object.

        All of an entity's one-to-one and one-to-many relationships are
        populated with objects of the correct type. For example, if
        ``SomeEntity.other_entity`` is a one-to-one relationship, this should
        return ``True``::

            isinstance(
                SomeEntity(id=N).read().other_entity,
                nailgun.entity_mixins.Entity
            )

        Additionally, both of these commands should succeed::

            SomeEntity(id=N).read().other_entity.id
            SomeEntity(id=N).read().other_entity.read().other_attr

        In the example above, ``other_entity.id`` is the **only** attribute
        with a meaningful value. Calling ``other_entity.read`` populates the
        remaining entity attributes.

        :param nailgun.entity_mixins.Entity entity: The object to be populated
            and returned. An object of type ``type(self)`` by default.
        :param attrs: A dict. Data used to populate the object's attributes.
            The response from
            :meth:`nailgun.entity_mixins.EntityReadMixin.read_json` by default.
        :param ignore: A tuple of attributes which should not be read from the
            server. This is mainly useful for attributes like a password which
            are not returned.
        :return: An instance of type ``type(self)``.
        :rtype: nailgun.entity_mixins.Entity

        """
        if entity is None:
            entity = type(self)(self._server_config)
        if attrs is None:
            attrs = self.read_json()

        # Rename fields using entity.Meta.api_names, if present.
        if hasattr(entity.Meta, 'api_names'):
            for local_name, remote_name in entity.Meta.api_names.items():
                attrs[local_name] = attrs.pop(remote_name)

        # The server returns OneToOneFields as a hash of attributes, like so:
        #
        #     {'name': 'Alice Hayworth', 'login': 'ahayworth', 'id': 1}
        #
        # OneToManyFields are a list of hashes.
        for field_name, field_type in entity.get_fields().items():
            if field_name in ignore:
                continue
            if isinstance(field_type, OneToOneField):
                if attrs[field_name] is None:
                    setattr(entity, field_name, None)
                else:
                    referenced_entity = _get_class(
                        field_type.entity,
                        'robottelo.entities'
                    )(self._server_config, id=attrs[field_name]['id'])
                    setattr(entity, field_name, referenced_entity)
            elif isinstance(field_type, OneToManyField):
                other_cls = _get_class(field_type.entity, 'robottelo.entities')
                referenced_entities = [
                    other_cls(self._server_config, id=referenced_entity['id'])
                    for referenced_entity
                    in attrs[field_name + 's']  # e.g. "users"
                ]
                setattr(entity, field_name, referenced_entities)
            else:
                setattr(entity, field_name, attrs[field_name])
        return entity


class EntityCreateMixin(object):
    """A mixin that provides the ability to create an entity.

    The methods provided by this mixin work together to create an entity. A
    typical tree of method calls looks like this::

        create
        └── create_json
            └── create_raw
                ├── create_missing
                └── create_payload

    Only :meth:`create_raw` communicates with the server.

    """

    def create_missing(self):
        """Automagically populate all required instance attributes.

        Iterate through the set of all required class ``Field`` defined on
        ``type(self)`` and create a corresponding instance attribute if none
        exists. Subclasses should override this method if there is some
        relationship between two required fields.

        :return: Nothing. This method relies on side-effects.

        """
        for field_name, field in self.get_fields().items():
            if field.required and field_name not in vars(self):
                # Most `gen_value` methods return a value such as an integer,
                # string or dictionary, but OneTo{One,Many}Field.gen_value
                # returns an instance of the referenced class.
                #
                # When populating a foreign key field, this is inadvisable:
                #
                #     value = entity_obj = field.gen_value()
                #     entity_obj.id = entity_obj.create_json()['id']
                #
                # The problem is that the values that entity_obj populates
                # itself with (via `create_missing`) may be different from the
                # values that are read back from the server, e.g. `Host.name`.
                if hasattr(field, 'default'):
                    value = field.default
                elif hasattr(field, 'choices'):
                    value = gen_choice(field.choices)
                elif isinstance(field, OneToOneField):
                    value = field.gen_value()
                    value.id = field.gen_value().create_json()['id']
                elif isinstance(field, OneToManyField):
                    value = [field.gen_value()]
                    value[0].id = field.gen_value().create_json()['id']
                else:
                    value = field.gen_value()
                setattr(self, field_name, value)

    def create_payload(self):
        """Create a payload that can be POSTed to the server.

        Make a copy of the instance attributes on ``self``. This payload will
        be POSTed to the server. Then change the payload's keys. (The payload
        is a dict.) Rename keys using ``self.Meta.api_names`` if it is present
        and append "_id" and "_ids" as appropriate.

        :return: A dict. A copy of the instance attributes on ``self``, with
            key names adjusted as appropriate.

        """
        data = self.get_values().copy()
        api_names = getattr(self.Meta, 'api_names', {})
        for field_name, field in type(self).get_fields().items():
            if field_name in data:
                if field_name in api_names:
                    # e.g. rename filter_type to type for ContentViewFilter
                    data[api_names[field_name]] = data.pop(field_name)
                if isinstance(field, OneToOneField):
                    data[field_name + '_id'] = data.pop(field_name).id
                elif isinstance(field, OneToManyField):
                    data[field_name + '_ids'] = [
                        entity.id for entity in data.pop(field_name)
                    ]
        return data

    def create_raw(self, create_missing=True):
        """Create an entity.

        Generate values for required, unset fields by calling
        :meth:`create_missing`. Only do this if ``create_missing`` is true.
        Then make an HTTP POST call to ``self.path('base')``. Return the
        response received from the server.

        :param create_missing: A boolean. Should :meth:`create_missing` be
            called? In other words, should values be generated for required,
            empty fields?
        :return: A ``requests.response`` object.

        """
        if create_missing:
            self.create_missing()
        return client.post(
            self.path('base'),
            self.create_payload(),
            auth=self._server_config.auth,
            verify=self._server_config.verify,
        )

    def create_json(self, create_missing=True):
        """Create an entity.

        Call :meth:`create_raw`. Check the response status code, decode JSON
        and return the decoded JSON as a dict.

        :return: A dict. The server's response, with all JSON decoded.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If the response JSON can not be decoded.

        """
        response = self.create_raw(create_missing)
        response.raise_for_status()
        return response.json()

    def create(self, create_missing=True):
        """Call :meth:`create_json`.

        This method exists for compatibility. It should be rewritten to match
        to act like :meth:`EntityReadMixin.read` after existing code is changed
        to use :meth:`create_json`.

        """
        return self.create_json(create_missing)



"""Tests for :mod:`nailgun.entity_mixins`."""
# (Too many public methods) pylint: disable=R0904
#
# Python 3.3 and later includes module `ipaddress` in the standard library. If
# Robottelo ever moves past Python 2.x, that module should be used instead of
# `socket`.
from fauxfactory import gen_integer
from nailgun import client, config, entity_mixins
from nailgun.entity_fields import (
    IntegerField,
    OneToManyField,
    OneToOneField,
    StringField,
)
import mock
import unittest


class SampleEntity(entity_mixins.Entity):
    """Sample entity to be used in the tests"""
    # pylint:disable=too-few-public-methods
    name = StringField()
    value = IntegerField()

    class Meta(object):
        """Non-field attributes for this entity."""
        # pylint:disable=too-few-public-methods
        api_path = 'foo'


class ManyRelatedEntity(entity_mixins.Entity):
    """An entity with a :class:`nailgun.entity_fields.OneToManyField`.

    The ``OneToOneField`` on this entity points to
    :class:`tests.test_entity_mixins.SampleEntity`.

    """
    # pylint:disable=too-few-public-methods
    entities = OneToManyField(SampleEntity)


class EntityWithDelete(entity_mixins.Entity, entity_mixins.EntityDeleteMixin):
    """
    An entity inheriting from :class:`nailgun.entity_mixins.EntityDeleteMixin`.
    """
    # pylint:disable=too-few-public-methods

    class Meta(object):
        """Non-field attributes for this entity."""
        # pylint:disable=too-few-public-methods
        api_path = ''


class EntityWithRead(entity_mixins.Entity, entity_mixins.EntityReadMixin):
    """
    An entity inheriting from :class:`nailgun.entity_mixins.EntityReadMixin`.
    """
    # pylint:disable=too-few-public-methods
    one_to_one = OneToOneField(SampleEntity)
    one_to_many = OneToManyField(SampleEntity)

    class Meta(object):
        """Non-field attributes for this entity."""
        # pylint:disable=too-few-public-methods
        api_path = ''


# -----------------------------------------------------------------------------


class MakeEntityFromIdTestCase(unittest.TestCase):
    """Tests for :func:`nailgun.entity_mixins._make_entity_from_id`."""
    # pylint:disable=protected-access

    def setUp(self):  # noqa pylint:disable=C0103
        """Set ``self.server_config``."""
        self.server_config = config.ServerConfig('example.com')

    def test_pass_in_entity_obj(self):
        """Pass in an entity class and an entity."""
        self.assertIsInstance(
            entity_mixins._make_entity_from_id(
                SampleEntity,
                SampleEntity(self.server_config),
                self.server_config
            ),
            SampleEntity
        )

    def test_pass_in_entity_id(self):
        """Pass in an entity class and an entity ID."""
        entity_id = gen_integer(min_value=1)
        entity_obj = entity_mixins._make_entity_from_id(
            SampleEntity,
            entity_id,
            self.server_config
        )
        self.assertIsInstance(entity_obj, SampleEntity)
        self.assertEqual(entity_obj.id, entity_id)


class MakeEntitiesFromIdsTestCase(unittest.TestCase):
    """Tests for ``_make_entity_from_ids``."""
    # pylint:disable=protected-access

    def setUp(self):  # noqa pylint:disable=C0103
        """Set ``self.server_config``."""
        self.server_config = config.ServerConfig('example.com')

    def test_pass_in_emtpy_iterable(self):
        """Pass in an entity class and an empty iterable."""
        for iterable in ([], tuple()):
            self.assertEqual(
                [],
                entity_mixins._make_entities_from_ids(
                    SampleEntity,
                    iterable,
                    self.server_config
                )
            )

    def test_pass_in_entity_obj(self):
        """Pass in an entity class and an iterable containing entities."""
        for num_entities in range(4):
            input_entities = [
                SampleEntity(self.server_config) for _ in range(num_entities)
            ]
            output_entities = entity_mixins._make_entities_from_ids(
                SampleEntity,
                input_entities,
                self.server_config
            )
            self.assertEqual(num_entities, len(output_entities))
            for output_entity in output_entities:
                self.assertIsInstance(output_entity, SampleEntity)

    def test_pass_in_entity_ids(self):
        """Pass in an entity class and an iterable containing entity IDs."""
        for num_entities in range(4):
            entity_ids = [
                gen_integer(min_value=1) for _ in range(num_entities)
            ]
            entities = entity_mixins._make_entities_from_ids(
                SampleEntity,
                entity_ids,
                self.server_config
            )
            self.assertEqual(len(entities), len(entity_ids))
            for i in range(len(entity_ids)):
                self.assertIsInstance(entities[i], SampleEntity)
                self.assertEqual(entities[i].id, entity_ids[i])

    def test_pass_in_both(self):
        """Pass in an entity class and an iterable with entities and IDs."""
        entities = entity_mixins._make_entities_from_ids(
            SampleEntity,
            [SampleEntity(self.server_config), 5],
            self.server_config
        )
        self.assertEqual(len(entities), 2)
        for entity in entities:
            self.assertIsInstance(entity, SampleEntity)


class EntityTestCase(unittest.TestCase):
    """Tests for :class:`nailgun.entity_mixins.Entity`."""

    def setUp(self):  # noqa pylint:disable=C0103
        """Set ``self.server_config`` and ``self.base_path``."""
        self.server_config = config.ServerConfig('http://example.com')
        self.base_path = '{0}/{1}'.format(
            self.server_config.url,
            SampleEntity.Meta.api_path
        )

    def test_entity_get_fields(self):
        """Test :meth:`nailgun.entity_mixins.Entity.get_fields`."""
        fields = SampleEntity(self.server_config).get_fields()

        self.assertIn('name', fields)
        self.assertIn('value', fields)

        self.assertIsInstance(fields['name'], StringField)
        self.assertIsInstance(fields['value'], IntegerField)

    def test_path(self):
        """Test :meth:`nailgun.entity_mixins.Entity.path`."""
        self.assertEqual(
            SampleEntity(self.server_config).path(),
            self.base_path,
        )
        self.assertEqual(
            SampleEntity(self.server_config, id=5).path(),
            self.base_path + '/5',
        )
        self.assertEqual(
            SampleEntity(self.server_config, id=5).path('base'),
            self.base_path,
        )
        with self.assertRaises(entity_mixins.NoSuchPathError):
            SampleEntity(self.server_config).path('self')


class EntityDeleteMixinTestCase(unittest.TestCase):
    """Tests for entity mixin classes."""

    def setUp(self):  # noqa pylint:disable=C0103
        """Set ``self.server_config`` and ``self.entity_id``."""
        # Example usage: SomeEntity(server_config, id=self.entity_id)
        self.server_config = config.ServerConfig(
            'example.com',
            auth=('Alice', 'hackme'),
            verify=True,
        )
        self.entity_id = gen_integer(min_value=1)

    def test_delete_200(self):
        """Test :meth:`nailgun.entity_mixins.EntityDeleteMixin.delete`.

        Assert that ``delete`` returns a JSON-decoded response from the server
        when an HTTP 200 status code is returned.

        """
        delete_return = mock.Mock()
        delete_return.status_code = 200
        delete_return.json.return_value = {'9501743': '14697417'}  # arbitrary
        with mock.patch.object(client, 'delete') as client_delete:
            client_delete.return_value = delete_return
            self.assertEqual(
                EntityWithDelete(
                    self.server_config,
                    id=self.entity_id
                ).delete(),
                delete_return.json.return_value
            )

    def test_delete_202(self):
        """Test :meth:`nailgun.entity_mixins.EntityDeleteMixin.delete`.

        Assert that ``delete`` returns information about a foreman task when an
        HTTP 202 status code is returned.

        """
        delete_return = mock.Mock()
        delete_return.status_code = 202
        delete_return.json.return_value = {'id': gen_integer()}  # mock task ID
        with mock.patch.object(client, 'delete') as client_delete:
            client_delete.return_value = delete_return
            with mock.patch.object(entity_mixins, '_poll_task') as poller:
                poller.return_value = {'324171': '59601212'}  # arbitrary
                self.assertEqual(
                    EntityWithDelete(
                        self.server_config,
                        id=self.entity_id
                    ).delete(),
                    poller.return_value
                )

    def test_read(self):
        """Test :meth:`nailgun.entity_mixins.EntityReadMixin.read`  and
        :meth:`nailgun.entity_mixins.EntityReadMixin.read_json`.

        Assert that ``EntityReadMixin.read_json`` returns the server's
        response, with all JSON decoded.

        Assert that ``EntityReadMixin.read`` returns an object with correctly
        populated attributes.

        """
        # Create a mock server response object.
        mock_response = mock.Mock()
        mock_response.raise_for_status.return_value = None
        mock_response.json.return_value = {
            u'id': self.entity_id,
            u'one_to_one': {'id': 123},
            u'one_to_manys': [{'id': 234}, {'id': 345}],
        }

        # Make `client.get` return the above object.
        with mock.patch.object(client, 'get') as client_get:
            client_get.return_value = mock_response

            # See if EntityReadMixin.read_json behaves correctly.
            self.assertEqual(
                EntityWithRead(
                    self.server_config,
                    id=self.entity_id
                ).read_json(),
                mock_response.json.return_value,
            )

            # See if EntityReadMixin.read behaves correctly.
            entity = EntityWithRead(
                self.server_config,
                id=self.entity_id
            ).read()
            self.assertEqual(entity.id, self.entity_id)
            self.assertIsInstance(entity.one_to_one, SampleEntity)
            self.assertEqual(entity.one_to_one.id, 123)
            self.assertEqual(len(entity.one_to_many), 2)
            for entity_ in entity.one_to_many:
                self.assertIsInstance(entity_, SampleEntity)
                self.assertIn(entity_.id, [234, 345])



"""Unit tests for :mod:`nailgun.config`."""
from mock import call, mock_open, patch
from nailgun.config import BaseServerConfig, ServerConfig
from unittest import TestCase
import json

from sys import version_info
if version_info[0] == 2:
    # The `__builtins__` module (note the "s") also provides the `open`
    # function. However, that module is an implementation detail for CPython 2,
    # so it should not be relied on.
    import __builtin__ as builtins  # pylint:disable=import-error
else:
    import builtins  # pylint:disable=import-error


FILE_PATH = '/tmp/bogus.json'
CONFIGS = {
    'default': {'url': 'http://example.com'},
    'Ask Aak': {'url': 'bogus value', 'auth': ['username', 'password']},
}
CONFIGS2 = CONFIGS.copy()
CONFIGS2.update({
    'Abeloth': {'url': 'bogus value', 'verify': True},
    'Admiral Gial Ackbar': {'url': 'bogus', 'auth': [], 'verify': False},
})


class BaseServerConfigTestCase(TestCase):
    """Tests for :class:`nailgun.config.BaseServerConfig`."""

    def test_init(self):
        """Test instantiating :class:`nailgun.config.BaseServerConfig`.

        Assert that only provided values become object attributes.

        """
        for config in CONFIGS.values():
            self.assertEqual(config, vars(BaseServerConfig(**config)))

    def test_get(self):
        """Test :meth:`nailgun.config.BaseServerConfig.get`.

        Assert that the method extracts the asked-for section from a
        configuration file and correctly populates a new ``BaseServerConfig``
        object. Also assert that the ``auth`` attribute is a list. (See the
        docstring for :meth:`nailgun.config.ServerConfig.get`.)

        """
        for label, config in CONFIGS.items():
            open_ = mock_open(read_data=json.dumps(CONFIGS))
            with patch.object(builtins, 'open', open_):
                server_config = BaseServerConfig.get(label, FILE_PATH)
            self.assertEqual(vars(server_config), config)
            open_.assert_called_once_with(FILE_PATH)
            if hasattr(server_config, 'auth'):
                self.assertIsInstance(server_config.auth, list)

    def test_get_labels(self):
        """Test :meth:`nailgun.config.BaseServerConfig.get_labels`.

        Assert that the method returns the correct labels.

        """
        open_ = mock_open(read_data=json.dumps(CONFIGS))
        with patch.object(builtins, 'open', open_):
            self.assertEqual(
                set(CONFIGS.keys()),
                set(BaseServerConfig.get_labels(FILE_PATH)),
            )
        open_.assert_called_once_with(FILE_PATH)

    def test_save(self):
        """Test :meth:`nailgun.config.BaseServerConfig.save`.

        Assert that the method reads the config file before writing, and that
        it writes out a correct config file.

        """
        label = 'Ask Aak'
        config = {label: {'url': 'https://example.org'}}
        open_ = mock_open(read_data=json.dumps(CONFIGS))
        with patch.object(builtins, 'open', open_):
            BaseServerConfig(config[label]['url']).save(label, FILE_PATH)

        # We care about two things: that this method reads the config file
        # before writing, and that the written string is correct. The first is
        # easy to verify...
        self.assertEqual(
            open_.call_args_list,
            [call(FILE_PATH), call(FILE_PATH, 'w')],
        )
        # ...and the second is a PITA to verify.
        actual_config = _get_written_json(open_)
        target_config = CONFIGS.copy()
        target_config[label] = config[label]
        self.assertEqual(target_config, actual_config)

    def test_delete(self):
        """Test :meth:`nailgun.config.BaseServerConfig.delete`.

        Assert that the method reads the config file before writing, and that
        it writes out a correct config file.

        """
        open_ = mock_open(read_data=json.dumps(CONFIGS))
        with patch.object(builtins, 'open', open_):
            BaseServerConfig.delete('Ask Aak', FILE_PATH)

        # See `test_save` for further commentary on what's being done here.
        self.assertEqual(
            open_.call_args_list,
            [call(FILE_PATH), call(FILE_PATH, 'w')],
        )
        actual_config = _get_written_json(open_)
        target_config = CONFIGS.copy()
        del target_config['Ask Aak']
        self.assertEqual(target_config, actual_config)


class ServerConfigTestCase(TestCase):
    """Tests for :class:`nailgun.config.ServerConfig`."""

    def test_init(self):
        """Test instantiating :class:`nailgun.config.ServerConfig`.

        Assert that only provided values become object attributes.

        """
        for config in CONFIGS2.values():
            self.assertEqual(config, vars(ServerConfig(**config)))

    def test_get_client_kwargs(self):
        """Test :meth:`nailgun.config.ServerConfig.get_client_kwargs`.

        Assert that:

        * ``get_client_kwargs`` returns all of the instance attributes from its
          object except the "url" attribute, and
        * no instance attributes from the object are removed.

        """
        for config in CONFIGS2.values():
            target = config.copy()
            target.pop('url')
            server_config = ServerConfig(**config)
            self.assertDictEqual(target, server_config.get_client_kwargs())
            self.assertDictEqual(
                vars(ServerConfig(**config)),
                vars(server_config)
            )

    def test_get(self):
        """Test :meth:`nailgun.config.ServerConfig.get`.

        Assert that the ``auth`` attribute is a tuple.

        """
        for label in CONFIGS.keys():
            open_ = mock_open(read_data=json.dumps(CONFIGS))
            with patch.object(builtins, 'open', open_):
                server_config = ServerConfig.get(label, FILE_PATH)
            if hasattr(server_config, 'auth'):
                self.assertIsInstance(server_config.auth, tuple)


def _get_written_json(mock_obj):
    """Return the JSON that has been written to a mock `open` object."""
    # json.dump() calls write() for each individual JSON token.
    return json.loads(''.join(
        tuple(call_obj)[1][0]
        for call_obj
        in mock_obj().write.mock_calls
    ))



"""Tools for managing and presenting server connection configurations.

NailGun needs to know certain facts about the remote server in order to do
anything useful. For example, NailGun needs to know the URL of the remote
server (e.g. 'https://example.com:250') and how to authenticate with the remote
server. :class:`nailgun.config.ServerConfig` eases the task of managing and
presenting that information.

"""
from os.path import isfile, join
from threading import Lock
from xdg import BaseDirectory
import json


class ConfigFileError(Exception):
    """Indicates an error occurred when locating a configuration file.

    .. WARNING:: This class will likely be moved to a separate Python package
        in a future release of NailGun. Be careful about making references to
        this class, as those references will likely need to be changed.

    """


def _get_config_file_path(xdg_config_dir, xdg_config_file):
    """Search ``XDG_CONFIG_DIRS`` for a config file and return the first found.

    Search each of the standard XDG configuration directories for a
    configuration file. Return as soon as a configuration file is found. Beware
    that by the time client code attempts to open the file, it may be gone or
    otherwise inaccessible.

    :param xdg_config_dir: A string. The name of the directory that is suffixed
        to the end of each of the ``XDG_CONFIG_DIRS`` paths.
    :param xdg_config_file: A string. The name of the configuration file that
        is being searched for.
    :returns: A ``str`` path to a configuration file.
    :raises nailgun.config.ConfigFileError: When no configuration file can be
        found.

    """
    for config_dir in BaseDirectory.load_config_paths(xdg_config_dir):
        path = join(config_dir, xdg_config_file)
        if isfile(path):
            return path
    raise ConfigFileError(
        'No configuration files could be located after searching for a file '
        'named "{0}" in the standard XDG configuration paths, such as '
        '"~/.config/{1}/".'.format(xdg_config_file, xdg_config_dir)
    )


class BaseServerConfig(object):
    """A minimal set of facts for communicating with a Satellite server.

    :param url: A string. The URL of a server. For example:
        `'https://example.com:250'`.
    :param auth: Credentials to use when communicating with the server. For
        example: `('username', 'password')`. No instance attribute is created
        if no value is provided.

    .. WARNING:: This class will likely be moved to a separate Python package
        in a future release of NailGun. Be careful about making references to
        this class, as those references will likely need to be changed.

    """
    # Used to lock access to the configuration file when performing certain
    # operations, such as saving.
    _file_lock = Lock()
    # The name of the directory appended to ``XDG_CONFIG_DIRS``.
    _xdg_config_dir = 'librobottelo'
    # The name of the file in which settings are stored.
    _xdg_config_file = 'settings.json'

    def __init__(self, url, auth=None):
        self.url = url
        if auth is not None:
            self.auth = auth

    @classmethod
    def delete(cls, label='default', path=None):
        """Delete a server configuration.

        This method is thread safe.

        :param label: A string. The configuration identified by ``label`` is
            deleted.
        :param path: A string. The configuration file to be manipulated.
            Defaults to what is returned by
            :func:`nailgun.config._get_config_file_path`.
        :returns: ``None``

        """
        if path is None:
            path = _get_config_file_path(
                cls._xdg_config_dir,
                cls._xdg_config_file
            )
        cls._file_lock.acquire()
        try:
            with open(path) as config_file:
                config = json.load(config_file)
            del config[label]
            with open(path, 'w') as config_file:
                json.dump(config, config_file)
        finally:
            cls._file_lock.release()

    @classmethod
    def get(cls, label='default', path=None):
        """Get a server configuration.

        :param label: A string. The configuration identified by ``label`` is
            read.
        :param path: A string. The configuration file to be manipulated.
            Defaults to what is returned by
            :func:`nailgun.config._get_config_file_path`.
        :returns: A brand new :class:`nailgun.config.ServerConfig` object whose
            attributes have been populated as appropriate.
        :rtype: ServerConfig

        """
        if path is None:
            path = _get_config_file_path(
                cls._xdg_config_dir,
                cls._xdg_config_file
            )
        with open(path) as config_file:
            return ServerConfig(**json.load(config_file)[label])

    @classmethod
    def get_labels(cls, path=None):
        """Get all server configuration labels.

        :param path: A string. The configuration file to be manipulated.
            Defaults to what is returned by
            :func:`nailgun.config._get_config_file_path`.
        :returns: Server configuration labels, where each label is a string.

        """
        if path is None:
            path = _get_config_file_path(
                cls._xdg_config_dir,
                cls._xdg_config_file
            )
        with open(path) as config_file:
            # keys() returns a list in Python 2 and a view in Python 3.
            return tuple(json.load(config_file).keys())

    def save(self, label='default', path=None):
        """Save the current connection configuration to a file.

        This method is thread safe.

        :param label: A string. An identifier for the current configuration.
            This allows multiple configurations with unique labels to be saved
            in a single file. If a configuration identified by ``label``
            already exists in the destination configuration file, it is
            replaced.
        :param path: A string. The configuration file to be manipulated. By
            default, an XDG-compliant configuration file is used. A
            configuration file is created if one does not exist already.
        :returns: ``None``

        """
        if path is None:
            path = join(
                BaseDirectory.save_config_path(self._xdg_config_dir),
                self._xdg_config_file
            )
        self._file_lock.acquire()
        try:
            # Either read an existing config or make an empty one. Then update
            # the config and write it out.
            try:
                with open(path) as config_file:
                    config = json.load(config_file)
            except IOError:
                config = {}
            config[label] = vars(self)
            with open(path, 'w') as config_file:
                json.dump(config, config_file)
        finally:
            self._file_lock.release()


class ServerConfig(BaseServerConfig):
    """Facts for communicating with a server's API.

    This class inherits from :class:`nailgun.config.BaseServerConfig`. Other
    constructor parameters are documented there.

    :param verify: A boolean. Should SSL be verified when communicating with
        the server? No instance attribute is created if no value is provided.

    """
    # pylint:disable=too-few-public-methods
    # It's OK that this class has only one public method. This class is
    # intentionally small so that the parent class can be re-used.
    _xdg_config_dir = 'nailgun'
    _xdg_config_file = 'server_configs.json'

    def __init__(self, url, auth=None, verify=None):
        super(ServerConfig, self).__init__(url, auth)
        if verify is not None:
            self.verify = verify

    def get_client_kwargs(self):
        """Get a dict of instance attributes, but with "url" omitted.

        This method makes working with :mod:`nailgun.client` more pleasant.
        Code such as the following can be written::

            cfg = ServerConfig.get()
            client.get(cfg.url + '/api/v2', **cfg.get_client_kwargs())

        This method has been placed here to promote a better layering of
        responsibilities: this class knows all about :mod:`nailgun.client`, and
        :mod:`nailgun.client` knows nothing about this class.

        """
        config = vars(self).copy()
        config.pop('url')
        return config

    @classmethod
    def get(cls, label='default', path=None):
        """If ``auth`` is a two element list, convert it to a tuple.

        The entity classes rely on the requests library to be a transport
        mechanism. The methods provided by that library, such as ``get`` and
        ``post``, accept an ``auth`` argument. That argument must be a tuple:

            Auth tuple to enable Basic/Digest/Custom HTTP Auth.

        However, the JSON decoder does not recognize a tuple as a type, and
        represents sequences of elements as a tuple. Compensate for that.

        This override is done here, and not in the base class, because the base
        class may be extracted out into a separate library and used in other
        contexts. In those contexts, the presence of a list may not matter or
        may be desireable.

        """
        config = super(ServerConfig, cls).get(label, path)
        if hasattr(config, 'auth') and isinstance(config.auth, list):
            config.auth = tuple(config.auth)
        return config



# -*- coding: utf-8 -*-
"""The root of the NailGun namespace.

NailGun's modules are organized in to a tree of dependencies, where each module
only knows about the modules below it in the tree. (The ``nailgun.entities``
module is in the process of being ported over from robottelo.) They can be
visualized like this::

    nailgun.entities
    └── nailgun.entity_mixins
        ├── nailgun.entity_fields
        ├── nailgun.config
        └── nailgun.client

As an end user, you'll typically want to use the classes exposed by
``nailgun.entities``.

"""



#!/usr/bin/env python
"""A setuptools-based script for installing NailGun."""
from setuptools import find_packages, setup  # prefer setuptools over distutils


with open('README.rst') as handle:
    LONG_DESCRIPTION = handle.read()


setup(
    name='nailgun',
    version='0.5.1',  # Should be identical to the version in docs/conf.py!
    description='A library that facilitates easy usage of the Satellite 6 API',
    long_description=LONG_DESCRIPTION,
    url='https://github.com/SatelliteQE/nailgun',
    author='Jeremy Audet',
    author_email='ichimonji10@gmail.com',
    license='GPLv3',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        ('License :: OSI Approved :: GNU General Public License v3 or later '
         '(GPLv3+)'),
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    packages=find_packages(),
    install_requires=['requests', 'pyxdg'],
)



"""The basic components of the NailGun ORM.

Each of the fields in this module corresponds to some type of information that
Satellite tracks. When paired the classes in ``nailgun.entity_mixins``, it is
possible to represent the entities that Satellite manages. For example,
consider this abbreviated class definition::

    class User(
            entity_mixins.Entity,
            entity_mixins.EntityCreateMixin,
            entity_mixins.EntityDeleteMixin,
            entity_mixins.EntityReadMixin):
        entity_fields.login = StringField(
            length=(1, 100),
            required=True,
            str_type=('alpha', 'alphanumeric', 'cjk', 'latin1'),
        )
        entity_fields.admin = BooleanField(null=True)
        entity_fields.firstname = StringField(null=True, length=(1, 50))
        entity_fields.lastname = StringField(null=True, length=(1, 50))
        entity_fields.mail = EmailField(required=True)
        entity_fields.password = StringField(required=True)

The class represents a user account on a Satellite server. Each of the fields
represents some piece of information that is associated with a user account,
and the mixins provide useful methods.

Fields are intended to be used declaratively. You probably should not be
interacting with the field classes or their methods directly. Instead, they are
used by the various mixins. For example, method ``EntityReadMixin.read`` can be
used like this::

    user = User(id=5).read()

``EntityReadMixin.read`` creates a new ``User`` object and populates it. The
method knows how to deal with the data returned by the server because of the
fields on the ``User`` class.

A secondary use of fields is to generate random data. For example, you could
call ``User.login.gen_value()`` (implemented at :meth:`StringField.gen_value`)
to generate a random login. Beware that these methods strive to produce the
most outrageous values that are still legal, so they will often return nonsense
UTF-8 values, which is unpleasant to work with.

"""
from fauxfactory import (
    gen_boolean,
    gen_choice,
    gen_date,
    gen_datetime,
    gen_email,
    gen_integer,
    gen_ipaddr,
    gen_mac,
    gen_netmask,
    gen_string,
    gen_url,
)
from importlib import import_module
from inspect import isclass
import random
# pylint:disable=too-few-public-methods
# The classes in this module serve a declarative role. It is OK that they don't
# do much.
#
# Pylint warns that several of the `gen_value` methods do not make use of
# `self`. That warning is disabled where appropriate. The reason is that the
# `gen_value` methods are intended to be easily overridden in the entity
# classes that use them. Giving all `gen_value` methods the same signature
# makes this easier to do. A consistent method signature does no harm, and it
# reduces the cognitive load on other programmers.


# A sentinel object, used when `None` does not suffice.
_SENTINEL = object()

#: The default module that will be searched when an entity class needs to be
#: found. Used by at least :class:`OneToOneField` and `OneToManyField`.
ENTITIES_MODULE = 'nailgun.entities'


class Field(object):
    """Base class to implement other fields

    Record this field's attributes.

    :param required: A boolean. Determines whether a value must be submitted to
        the server when creating or updating an entity.
    :param choices: A tuple of values that this field may be populated with.
    :param default: Entity classes that inherit from ``EntityCreateMixin`` make
        use of this field.
    :param null: A boolean. Determines whether a null value can be submitted to
        the server when creating or updating an entity.

    """

    def __init__(
            self,
            required=False,
            choices=None,
            default=_SENTINEL,
            null=False):
        self.required = required
        if choices is not None:
            self.choices = choices
        if default is not _SENTINEL:
            self.default = default
        self.null = null


class BooleanField(Field):
    """Field that represents a boolean"""

    def gen_value(self):
        """Return a value suitable for a :class:`BooleanField`."""
        # pylint:disable=no-self-use
        # See comment near top of module.
        return gen_boolean()


class EmailField(Field):
    """Field that represents an email"""

    def gen_value(self):
        """Return a value suitable for a :class:`EmailField`."""
        # pylint:disable=no-self-use
        # See comment near top of module.
        return gen_email()


class FloatField(Field):
    """Field that represents a float"""

    def gen_value(self):
        """Return a value suitable for a :class:`FloatField`."""
        # pylint:disable=no-self-use
        # See comment near top of module.
        return random.random() * 10000


class IntegerField(Field):
    """Field that represents an integer."""

    def __init__(self, min_val=None, max_val=None, *args, **kwargs):
        self.min_val = min_val
        self.max_val = max_val
        super(IntegerField, self).__init__(*args, **kwargs)

    def gen_value(self):
        """Return a value suitable for a :class:`IntegerField`."""
        return gen_integer(self.min_val, self.max_val)


class StringField(Field):
    """Field that represents a string.

    The default ``length`` of string fields is short for two reasons:

    1. Foreman's database backend limits many fields to 255 bytes in length. As
       a result, ``length`` should be no longer than 85 characters long, as 85
       unicode characters may be up to 255 bytes long.
    2. Humans have to read through the error messages produced by Robottelo.
       Long error messages are hard to read through, and that hurts
       productivity. Thus, a ``length`` even shorter than 85 chars is
       desirable.

    :param length: Either a ``(min_len, max_len)`` tuple or an ``exact_len``
        integer.
    :param str_type: The types of characters to generate when
        :meth:`StringField.gen_value` is called. May be a single string type
        (e.g. ``'utf8'``) or a tuple of string types. This argument is passed
        through to FauxFactory's ``gen_string`` method, so this method accepts
        all string types which that method does.

    """

    def __init__(self, length=(1, 30), str_type=('utf8',), *args, **kwargs):
        # length may be a two-tuple or an integer. Set {min,max}_len carefully.
        if isinstance(length, tuple):
            self.min_len, self.max_len = length
        else:
            self.min_len = self.max_len = length

        # str_type may be a tuple or a string, but self.str_type is a tuple.
        if isinstance(str_type, tuple):
            self.str_type = str_type
        else:
            self.str_type = (str_type,)

        super(StringField, self).__init__(*args, **kwargs)

    def gen_value(self):
        """Return a value suitable for a :class:`StringField`."""
        return gen_string(
            gen_choice(self.str_type),
            gen_integer(self.min_len, self.max_len)
        )


class DateField(Field):
    """Field that represents a date"""

    def __init__(self, min_date=None, max_date=None, *args, **kwargs):
        # If ``None`` is passed then ``FauxFactory`` will deal with it.
        self.min_date = min_date
        self.max_date = max_date
        super(DateField, self).__init__(*args, **kwargs)

    def gen_value(self):
        """Return a value suitable for a :class:`DateField`."""
        return gen_date(self.min_date, self.max_date)


class DateTimeField(Field):
    """Field that represents a datetime"""

    def __init__(self, min_date=None, max_date=None, *args, **kwargs):
        # If ``None`` is passed then ``FauxFactory`` will deal with it.
        self.min_date = min_date
        self.max_date = max_date
        super(DateTimeField, self).__init__(*args, **kwargs)

    def gen_value(self):
        """Return a value suitable for a :class:`DateTimeField`."""
        return gen_datetime(self.min_date, self.max_date)


class DictField(Field):
    """Field that represents a set of key-value pairs."""

    def gen_value(self):
        """Return a value suitable for a :class:`DictField`."""
        # pylint:disable=no-self-use
        # See comment near top of module.
        return {}


class IPAddressField(StringField):
    """Field that represents an IP adrress"""

    def gen_value(self):
        """Return a value suitable for a :class:`IPAddressField`."""
        return gen_ipaddr()


class NetmaskField(StringField):
    """Field that represents an netmask"""

    def gen_value(self):
        """Return a value suitable for a :class:`NetmaskField`."""
        return gen_netmask()


class ListField(Field):
    """Field that represents a list of strings"""


class MACAddressField(StringField):
    """Field that represents a MAC adrress"""

    def gen_value(self):
        """Return a value suitable for a :class:`MACAddressField`."""
        return gen_mac()


class OneToOneField(Field):
    """Field that represents a reference to another entity.

    All parameters not documented here are passed to :class:`Field`.

    :param entity: Either a class or the name of a class.
    :param module: A string. A dotted module name. Used by
        :meth:`OneToOneField.gen_value`. See that method for more details.

    """

    def __init__(self, entity, module=None, *args, **kwargs):
        self.entity = entity
        self.module = module
        super(OneToOneField, self).__init__(*args, **kwargs)

    def gen_value(self):
        """Return an instance of the class that this field references.

        If ``self.entity`` is a class, return an instance of that class.

        If ``self.entity`` is a string, find a class by that name and return an
        instance of it:

        * By default, search :data:`ENTITIES_MODULE` for the class.
        * If ``self.module`` is a string (e.g. ``'robottelo.entities'``), then
          search that module instead.

        """
        return _get_class(self.entity, self.module)()


class OneToManyField(Field):
    """Field that represents a reference to zero or more other entities.

    The parameters for this class are exactly the same as for
    :class:`OneToOneField`.

    """

    def __init__(self, entity, module=None, *args, **kwargs):
        self.entity = entity
        self.module = module
        super(OneToManyField, self).__init__(*args, **kwargs)

    def gen_value(self):
        """Return an instance of the class that this field references.

        This method behaves exactly like :meth:`OneToOneField.gen_value`.

        """
        return _get_class(self.entity, self.module)()


class URLField(StringField):
    """Field that represents an URL"""

    def gen_value(self):
        """Return a value suitable for a :class:`URLField`."""
        return gen_url()


def _get_class(class_or_name, module=None):
    """Return a class object.

    If ``class_or_name`` is a class, it is returned untouched. Otherwise,
    ``class_or_name`` is assumed to be a string. In this case, ``module`` is
    searched for a class by that name and returned.

    :param class_or_name: Either a class or the name of a class.
    :param module: A string. A dotted module name.
    :return: Either the class passed in or a class from ``module``.

    """
    if isclass(class_or_name):
        return class_or_name
    if module is None:
        module = ENTITIES_MODULE
    return getattr(import_module(module), class_or_name)



"""Unit tests for :mod:`nailgun.entity_fields`."""
from nailgun import entity_fields
from random import randint
from sys import version_info
from unittest import TestCase
import datetime
import socket


# It is OK that this class has no public methods. It just needs to exist for use
# by other tests, not be independently useful.
class TestClass(object):  # pylint:disable=too-few-public-methods
    """A class that is used when testing the OneTo{One,Many}Field classes."""


class GenValueTestCase(TestCase):
    """Tests for the ``gen_value`` method on various ``*Field`` classes.

    Classes with complex ``gen_value`` implementations are broken out into
    separate test cases.

    """

    def test_one_to_one_field(self):
        """Test :class:`nailgun.entity_fields.OneToOneField`."""
        for obj in (
                entity_fields.OneToOneField(TestClass).gen_value(),
                entity_fields.OneToOneField(
                    'TestClass',
                    'tests.test_entity_fields'
                ).gen_value(),
        ):
            self.assertIsInstance(obj, TestClass)

    def test_one_to_many_field(self):
        """Test :class:`nailgun.entity_fields.OneToManyField`."""
        for obj in (
                entity_fields.OneToManyField(TestClass).gen_value(),
                entity_fields.OneToManyField(
                    'TestClass',
                    'tests.test_entity_fields'
                ).gen_value(),
        ):
            self.assertIsInstance(obj, TestClass)

    def test_boolean_field(self):
        """Test :class:`nailgun.entity_fields.BooleanField`."""
        self.assertIn(entity_fields.BooleanField().gen_value(), (True, False))

    def test_date_field(self):
        """Test :class:`nailgun.entity_fields.DateField`."""
        self.assertIsInstance(
            entity_fields.DateField().gen_value(), datetime.date
        )

    def test_datetime_field(self):
        """Test :class:`nailgun.entity_fields.DateTimeField`."""
        self.assertIsInstance(
            entity_fields.DateTimeField().gen_value(), datetime.datetime
        )

    def test_email_field(self):
        """Test :class:`nailgun.entity_fields.EmailField`.

        Ensure :meth:`nailgun.entity_fields.EmailField.gen_value` returns a
        unicode string containing the character '@'.

        """
        email = entity_fields.EmailField().gen_value()
        if version_info.major == 2:
            self.assertIsInstance(email, unicode)  # flake8:noqa pylint:disable=undefined-variable
        else:
            self.assertIsInstance(email, str)
        self.assertIn('@', email)

    def test_float_field(self):
        """Test :class:`nailgun.entity_fields.FloatField`."""
        self.assertIsInstance(entity_fields.FloatField().gen_value(), float)

    def test_ip_address_field(self):
        """Test :class:`nailgun.entity_fields.IPAddressField`.

        Ensure the value returned is acceptable to ``socket.inet_aton``.

        """
        addr = entity_fields.IPAddressField().gen_value()
        try:
            socket.inet_aton(addr)
        except socket.error as err:
            self.fail('({0}) {1}'.format(addr, err))

    def test_mac_address_field(self):
        """Test :class:`nailgun.entity_fields.MACAddressField`.

        Ensure the value returned is a string containing 12 hex digits (either
        upper or lower case), grouped into pairs of digits and separated by
        colon characters. For example: ``'01:23:45:FE:dc:BA'``

        The regex used in this test is inspired by this Q&A:
        http://stackoverflow.com/questions/7629643/how-do-i-validate-the-format-of-a-mac-address

        """
        if version_info.major == 2:
            validator = self.assertRegexpMatches
        else:
            validator = self.assertRegex  # pylint:disable=no-member
        validator(
            entity_fields.MACAddressField().gen_value().upper(),
            '^([0-9A-F]{2}[:]){5}[0-9A-F]{2}$'
        )


class StringFieldTestCase(TestCase):
    """Tests for :class:`nailgun.entity_fields.StringField`."""

    def test_str_is_returned(self):
        """Ensure a unicode string at least 1 char long is returned."""
        string = entity_fields.StringField().gen_value()
        if version_info[0] == 2:
            self.assertIsInstance(string, unicode)  # flake8:noqa pylint:disable=undefined-variable
        else:
            self.assertIsInstance(string, str)
        self.assertGreater(len(string), 0)

    def test_length_arg(self):
        """Ensure that the ``length`` argument is respected."""
        # What happens when we pass in an exact length?
        string = entity_fields.StringField(length=5).gen_value()
        self.assertEqual(len(string), 5)

        # What happens when we pass in a range of lengths?
        string = entity_fields.StringField(length=(1, 20)).gen_value()
        self.assertGreaterEqual(len(string), 1)
        self.assertLessEqual(len(string), 20)

    def test_str_type_arg(self):
        """Ensure that the ``str_type`` argument is respected."""
        # This method uses single-use variables. But the code just looks
        # ridiculous if they are eliminated.
        for str_type in ('alpha', ('alpha',)):
            string = entity_fields.StringField(str_type=str_type).gen_value()
            self.assertTrue(string.isalpha())
        for str_type in ('numeric', ('numeric',)):
            string = entity_fields.StringField(str_type=str_type).gen_value()
            self.assertTrue(string.isnumeric())

        str_type = ('alpha', 'numeric')
        string = entity_fields.StringField(str_type=str_type).gen_value()
        self.assertTrue(string.isalpha() or string.isnumeric())


class IntegerFieldTestCase(TestCase):
    """Tests for :class:`nailgun.entity_fields.IntegerField`."""

    def test_int_is_returned(self):
        """Enture the value returned is an ``int``."""
        self.assertIsInstance(entity_fields.IntegerField().gen_value(), int)

    def test_min_val_arg(self):
        """Ensure that the ``min_val`` argument is respected."""
        min_val = randint(-1000, 1000)
        val = entity_fields.IntegerField(min_val=min_val).gen_value()
        self.assertGreaterEqual(val, min_val)

    def test_max_val_arg(self):
        """Ensure that the ``max_val`` argument is respected."""
        max_val = randint(-1000, 1000)
        val = entity_fields.IntegerField(max_val=max_val).gen_value()
        self.assertLessEqual(val, max_val)

    def test_min_val_max_val_args(self):
        """Ensure that the ``min_val`` and ``max_val`` args are respected."""
        min_val = randint(-1000, 0)
        max_val = randint(0, 1000)

        # First, we'll allow a range of values...
        val = entity_fields.IntegerField(min_val, max_val).gen_value()
        self.assertGreaterEqual(val, min_val)
        self.assertLessEqual(val, max_val)

        # ... then, we'll allow only a single value...
        val = entity_fields.IntegerField(min_val, min_val).gen_value()
        self.assertEqual(val, min_val)

        # ... twice over, just to be sure.
        val = entity_fields.IntegerField(max_val, max_val).gen_value()
        self.assertEqual(val, max_val)
